#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#


def split(input=["X", "y"], config={"test_size": 0.33, "shuffle": True, "stratify": "y", "random_state": 0}, **kwargs):
    r"""
    >>> from idict import idict, let
    >>> d = idict.fromtoy() >> split
    >>> d.show(colored=False)
    {
        "Xtr": "→(input config X y)",
        "ytr": "→(input config X y)",
        "Xts": "→(input config X y)",
        "yts": "→(input config X y)",
        "_history": "split----------------------sklearn-1.0.1",
        "X": "«{'attr1': {0: 5.1, 1: 1.1, 2: 6.1, 3: 1.1, 4: 3.1, 5: 4.7, 6: 9.1, 7: 8.3, 8: 9.1, 9: 2.5, 10: 7.1, 11: 0.1, 12: 2.1, 13: 0.1, 14: 5.1, 15: 31.1, 16: 1.1, 17: 2.2, 18: 3.1, 19: 1.1}, 'attr2': {0: 6.4, 1: 2.5, 2: 3.6, 3: 3.5, 4: 2.5, 5: 4.9, 6: 3.5, 7: 2.9, 8: 7.2, 9: 4.5, 10: 6.6, 11: 4.3, 12: 0.1, 13: 4.0, 14: 4.5, 15: 4.7, 16: 3.2, 17: 8.5, 18: 2.5, 19: 8.5}}»",
        "y": "«[0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1]»",
        "_id": "6xSiHAhSJC9yzGXeabFahIqeniPsklearn-1.0.1",
        "_ids": {
            "Xtr": "rE3WM8cu5e1XrvmPdKRVr9DJ0-L9mlearj-1.0.2",
            "ytr": "buTGORt9MCqtqUmtRAaCUY1dot1tklearn-1.0.3",
            "Xts": "6Z42mf0rylBy0cJV0tvi7naILYOsklearn-1.0.4",
            "yts": "ewHi412m.1H0.kMKdlQ-tCib7s4tklearn-1.0.5",
            "_history": "Cx0YhJSEq21qM7lw-zjsLwiJ1PelT6lxKxAV2Z2M",
            "X": "-2_3606060c77dc8f0807deec66fac5120578d0e (content: 34_1738c83af436029507def2710bc5125f58d0e)",
            "y": "o._080bf2a35d8ab275d4050dfd02b939104feac (content: S0_b6360d62ccafa275d4051dfd02b939104feac)"
        }
    }
    >>> d.yts
    array([1, 0, 1, 0, 0, 0, 1])
    """
    if input != ["X", "y"]:
        # TODO create a way in ldict to accept a dynamic dict as return value
        raise Exception(f"Not implemented for input/output different from default values: " f"{input}")
    from sklearn.model_selection import train_test_split

    # Multidynamic input is only detected when the kwargs index is also indexed by something.
    args = {}
    for i, _ in enumerate(input):
        args[input[i]] = kwargs[input[i]]

    if "stratify" in config and isinstance(config["stratify"], str):
        if config["stratify"] not in input:  # pragma: no cover
            raise Exception(f"Missing field {config['stratify']} for stratification.")
        config["stratify"] = args[config["stratify"]]
    Xtr, Xts, ytr, yts = train_test_split(*args.values(), **config)
    return {"Xtr": Xtr, "ytr": ytr, "Xts": Xts, "yts": yts, "_history": ...}


split.metadata = {
    "id": "split----------------------sklearn-1.0.1",
    "name": "split",
    "description": "Split data in two sets.",
    "parameters": ...,
    "code": ...,
}

# TODO: enforce versions by runtime checking (sklearn etc)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#


def fit(algorithm=None, config={}, Xin="X", yin="y", output="model", version=0, **kwargs):
    """
    >>> from sklearn.ensemble import RandomForestClassifier as RF
    >>> from idict import idict, let
    >>> d = idict.fromtoy() >> let(fit, algorithm=RF)
    >>> d.model
    RandomForestClassifier()
    >>> d >>= predict
    >>> d.z
    array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    """
    obj = algorithm(**config)
    obj.fit(kwargs[Xin], kwargs[yin])
    return {output: obj, "_history": ...}


fit.metadata = {
    "id": "fit--------------------------------idict",
    "name": "fit",
    "description": "Induce a model.",
    "parameters": ...,
    "code": ...,
}


def predict(input="model", Xin="X", yout="z", version=0, **kwargs):
    return {yout: kwargs[input].predict(kwargs[Xin]), "_history": ...}


predict.metadata = {
    "id": "predict----------------------------idict",
    "name": "predict",
    "description": "Predict values according to a model.",
    "parameters": ...,
    "code": ...,
}



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from json import loads

from ldict.exception import check_package
from orjson import dumps, OPT_SORT_KEYS, OPT_SERIALIZE_NUMPY


def default_orjson_encoder(obj):
    r"""
    >>> import numpy as np
    >>> default_orjson_encoder(np.array([[1/3,5/4],[1.3**64, -1]]))
    {'_type': "<class 'numpy.ndarray'>", 'repr': 'str', 'dtype': 'float64', 'obj': '[[0.3333333333333333,1.25],[19605347.64307615,-1.0]]'}
    """
    if m := check_package("pandas.core", obj):
        if isinstance(obj, (m.frame.DataFrame, m.series.Series)):
            dic = obj.to_dict()
            dic["_type"] = str(type(obj))
            return dic
    elif m := check_package("numpy", obj):
        if isinstance(obj, m.ndarray):
            try:
                dump = dumps(obj, option=OPT_SORT_KEYS | OPT_SERIALIZE_NUMPY)
                obj2, repr = dump.decode(encoding="utf8"), "str"
            except TypeError:
                obj2, repr = obj.tolist(), "list"
            return {"_type": str(type(obj)), "repr": repr, "dtype": str(obj.dtype), "obj": obj2}
    raise TypeError  # pragma: no cover


def encode(obj):
    r"""
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     [[1/3, 5/4], [1.3**54, "text"]],
    ...     index=["row 1", "row 2"],
    ...     columns=["col 1", "col 2"],
    ... )
    >>> df
                  col 1 col 2
    row 1  3.333333e-01  1.25
    row 2  1.422136e+06  text
    >>> encode(df)
    b'{"_type":"<class \'pandas.core.frame.DataFrame\'>","col 1":{"row 1":0.3333333333333333,"row 2":1422135.6537506874},"col 2":{"row 1":1.25,"row 2":"text"}}'
    >>> s = pd.Series(
    ...     [1/3, 5/4, (1.3)**54, "text"],
    ...     index=["row 1", "row 2", "row 3", "row 4"],
    ... )
    >>> s
    row 1          0.333333
    row 2              1.25
    row 3    1422135.653751
    row 4              text
    dtype: object
    >>> encode(s)
    b'{"_type":"<class \'pandas.core.series.Series\'>","row 1":0.3333333333333333,"row 2":1.25,"row 3":1422135.6537506874,"row 4":"text"}'
    >>> import numpy as np
    >>> encode(np.array([[1/3,5/4],[1.3**64, "text"]]))
    b'{"_type":"<class \'numpy.ndarray\'>","dtype":"<U32","obj":[["0.3333333333333333","1.25"],["19605347.64307615","text"]],"repr":"list"}'
    """
    return dumps(obj, default=default_orjson_encoder, option=OPT_SORT_KEYS)


def json_object_hook_decoder(dic):
    if "_type" in dic:
        if "pandas" in (typ := dic.pop("_type")):
            m = check_package("pandas.core")
            if "DataFrame" in typ:
                return m.frame.DataFrame.from_dict(dic)
            elif "Series" in typ:
                return m.series.Series(data=dic.values(), index=dic.keys())
        if "ndarray" in typ and "numpy" in typ:
            m = check_package("numpy")
            obj = loads(dic["obj"]) if dic["repr"] == "str" else dic["obj"]
            return m.array(obj, dtype=dic["dtype"])
    return dic


def decode(blob):
    r"""
    >>> df = b'{"_type":"<class \'pandas.core.frame.DataFrame\'>","col 1":{"row 1":0.3333333333333333,"row 2":1422135.6537506874},"col 2":{"row 1":1.25,"row 2":"text"}}'
    >>> decode(df)
                  col 1 col 2
    row 1  3.333333e-01  1.25
    row 2  1.422136e+06  text
    >>> s = b'{"_type":"<class \'pandas.core.series.Series\'>","row 1":0.3333333333333333,"row 2":1.25,"row 3":1422135.6537506874,"row 4":"text"}'
    >>> decode(s)
    row 1          0.333333
    row 2              1.25
    row 3    1422135.653751
    row 4              text
    dtype: object
    >>> n = {'_type': "<class 'numpy.ndarray'>", 'repr': 'str', 'dtype': 'float64', 'obj': '[[0.3333333333333333,1.25],[19605347.64307615,-1.0]]'}
    >>> decode(encode(n))
    array([[ 3.33333333e-01,  1.25000000e+00],
           [ 1.96053476e+07, -1.00000000e+00]])
    """
    return loads(blob, object_hook=json_object_hook_decoder)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

"""
Functions to be used directly within an idict workflow
"""
from io import BytesIO, StringIO

from arff2pandas import a2p


def df2np(input="df", Xout="X", yout="y", **kwargs):
    """
    >>> from idict import let, idict
    >>> d = idict.fromminiarff()
    >>> d >>= df2np
    >>> d.show(colored=False)
    {
        "X": "→(input Xout yout df)",
        "y": "→(input Xout yout df)",
        "_history": "df2np--------pandas-1.3.4--sklearn-1.0.1",
        "df": "«{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}»",
        "_id": "E44sczWfUMWFuItAavOd9v.4Q1.sklearn-1.0.1",
        "_ids": {
            "X": "d4mcz9Xanx8fRsribw-L8-UdKxF9mlearj-1.0.2",
            "y": "fSwn3UKP-APakeSBnJMc6f5n6.6tklearn-1.0.3",
            "_history": "A3RPrmvmGNuH9IfL02DgQY2007VEzNNRA8xZ0.0X",
            "df": "ja_3dbc3e0089a672ae7896199398b692362dc99 (content: 6X_dc8ccea3b2e46f1c78967fae98b692701dc99)"
        }
    }
    >>> d.y
    array([0, 1])
    """
    from sklearn.preprocessing import LabelEncoder

    le = LabelEncoder()
    df = kwargs[input]
    X_ = df.drop(df.columns[[-1]], axis=1)
    y_ = le.fit_transform(df[df.columns[-1]])
    return {Xout: X_, yout: y_, "_history": ...}


df2np.metadata = {
    "id": "df2np--------pandas-1.3.4--sklearn-1.0.1",
    "name": "df2np",
    "description": "DataFrame (pandas) to X,y (numpy) converter.",
    "parameters": ...,
    "code": ...,
}


def df2arff(input="df", output="arff", **kwargs):
    """
    >>> from idict import let, idict
    >>> d = idict.fromminiarff()
    >>> d >>= let(df2arff, output="a")
    >>> d.show(colored=False)
    {
        "a": "→(input output df)",
        "_history": "df2arff----------------arff2pandas-1.0.1",
        "df": "«{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}»",
        "_id": "xdyZaF3yh4jf0fQK8IxSo0m5XKfLPerBLwVuTve8",
        "_ids": {
            "a": "ThqwG-KOhpzcRprusvXcZwjDrLgLPerBLwVuTve9",
            "_history": "ofEb.nRSYsUsgAnnyp4KYFovZaUOV6000sv....-",
            "df": "ja_3dbc3e0089a672ae7896199398b692362dc99 (content: 6X_dc8ccea3b2e46f1c78967fae98b692701dc99)"
        }
    }
    >>> d.a
    '@RELATION data\\n\\n@ATTRIBUTE attr1 REAL\\n@ATTRIBUTE attr2 REAL\\n@ATTRIBUTE class {0, 1}\\n\\n@DATA\\n5.1,3.5,0\\n3.1,4.5,1\\n'
    """
    from arff2pandas import a2p

    return {output: a2p.dumps(kwargs[input]), "_history": ...}


df2arff.metadata = {
    "id": "df2arff----------------arff2pandas-1.0.1",
    "name": "df2arff",
    "description": "DataFrame (pandas) to ARFF converter.",
    "parameters": ...,
    "code": ...,
}


def openml(Xout="X", yout="y", name="iris", version=1):
    """
    >>> from idict import Ø
    >>> (Ø >> openml).show(colored=False)
    {
        "X": "→(Xout yout name version)",
        "y": "→(Xout yout name version)",
        "_history": "openml---------------------sklearn-1.0.1",
        "_id": "openml---------------------sklearn-1.0.1",
        "_ids": {
            "X": "ze11cyj1Gi2mcWmsh1rrjkjdzgI9mlearj-1.0.2",
            "y": "nw5qXzxIkHrUajn6VTL7M7KIWLZsklearn-1.0.3",
            "_history": "W46Q.9OBxnS9oeFTzDYkRbWSoDQEzNNRA8xZ0.0X"
        }
    }
    >>> (Ø >> openml).X.head()
       sepallength  sepalwidth  petallength  petalwidth
    0          5.1         3.5          1.4         0.2
    1          4.9         3.0          1.4         0.2
    2          4.7         3.2          1.3         0.2
    3          4.6         3.1          1.5         0.2
    4          5.0         3.6          1.4         0.2
    >>> (Ø >> openml).y.head()
    0    Iris-setosa
    1    Iris-setosa
    2    Iris-setosa
    3    Iris-setosa
    4    Iris-setosa
    Name: class, dtype: category
    Categories (3, object): ['Iris-setosa', 'Iris-versicolor', 'Iris-virginica']
    """
    from sklearn.datasets import fetch_openml

    X, y = fetch_openml(name=name, version=version, as_frame=True, return_X_y=True)
    return {Xout: X, yout: y, "_history": ...}


openml.metadata = {
    "id": "openml---------------------sklearn-1.0.1",
    "name": "openml",
    "description": "Fetch DataFrame+Series (pandas) from OpenML.",
    "parameters": ...,
    "code": ...,
}


# todo-tentar criar xy de DF usando x=DF e y=series, em vez de numpy. testar com RF      df[df.columns[-1]]


def arff2df(input="arff", output="df", **kwargs):
    r"""
    >>> from idict import let, idict
    >>> d = idict.fromminiarff(output=["arff"], output_format="arff")
    >>> d.arff
    '@RELATION mini\n@ATTRIBUTE attr1\tREAL\n@ATTRIBUTE attr2 \tREAL\n@ATTRIBUTE class \t{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1'
    >>> d >>= arff2df
    >>> d.show(colored=False)
    {
        "df": "→(input output arff)",
        "_history": "arff2df----------------arff2pandas-1.0.1",
        "arff": "@RELATION mini\n@ATTRIBUTE attr1\tREAL\n@ATTRIBUTE attr2 \tREAL\n@ATTRIBUTE class \t{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1",
        "_id": "n4EMpHwlrvpM-fFo4LC5914xb892pandas-1.0.1",
        "_ids": {
            "df": "ro3YIlxbNkgwanHshPzTghlh3T.Kqandao-1.0.2",
            "_history": "ofEb.nRSYsUsgAnnyp4KYFovZaUOV6000sv....-",
            "arff": "OW_29bd5266cf0a6400c5747b4c332b4a54d1955 (content: Ev_8bb973161e5ae900c5743b3c332b4a64d1955)"
        }
    }
    >>> d.df
       attr1@REAL  attr2@REAL class@{0,1}
    0         5.1         3.5           0
    1         3.1         4.5           1
    """
    with StringIO() as f:
        f.write(kwargs[input])
        df = a2p.loads(f.getvalue())

    return {output: df, "_history": ...}


arff2df.metadata = {
    "id": "arff2df----------------arff2pandas-1.0.1",
    "name": "arff2df",
    "description": "ARFF to DataFrame (pandas) converter.",
    "parameters": ...,
    "code": ...,
}
